/* outbreak.c
 * 
 * Copyright 2012 Hajo Möller <mail@dasjoe.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 * 
 */

#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>
#include <inttypes.h>
#include <limits.h>
#include <math.h>

#include "outbreak.h"

#define INDEXCASES 100

//~ typedef struct s_person {
	//~ uint8_t health;
	//~ uint8_t age;
	//~ uint8_t infected; /* disease_1 & disease_2 & ... */
	//~ uint8_t immune; /* immunity_1 & immunity_2 & ... */
	//~ uint8_t *incubation;
	//~ uint8_t *latency;
//~ } PERSON;

typedef struct s_area {
	uint8_t type; /* 0: empty, 1: water, else: land */
	float coeff_age;
	float coeff_reproduction;
	uint64_t population;
	uint64_t infected;
	uint64_t alive;
	uint64_t dead;
	//~ PERSON *people;
} AREA;

AREA **world;
AREA *w;
uint8_t w1 = 0;
uint8_t w2 = 1;

uint32_t frame = 0;

uint32_t a_width = 200;
uint32_t a_height = 200;
uint8_t a_countries;
double_t a_watertoland;
uint32_t a_population;
double_t a_travelchance;
double_t a_contagiousness;
double_t a_deadliness;
double_t a_curability;

int parseargs(int argc, char **argv) {
	if(argc != 10) {
		printf("Outbreak (%s, compiled %s %s)\n\n", VERSION, __DATE__, __TIME__);
		printf("Usage: outbreak [width] [height] [countries] [watertoland] [population] [travelchance] [contagiousness] [deadliness] [curability]\n\n");
		printf("  [width]           (uint32_t) width of simulation\n");
		printf("  [height]          (uint32_t) height of simulation\n");
		printf("  [countries]       (uint8_t)  countries to spawn\n");
		printf("  [watertoland]     (double_t) water / land\n");
		printf("  [population]      (uint32_t) count of people per area\n");
		printf("  [travelchance]    (double_t) chance any given person will travel this round\n");
		printf("  [contagiousness]  (double_t) people infected by 1 sick person\n");
		printf("  [deadliness]      (double_t) chance of a sick person dying\n");
		printf("  [curability]      (double_t) chance of spontaneous healing\n");
		return 1;
	} else {
		sscanf(argv[1], "%" SCNu32, &a_width);
		sscanf(argv[2], "%" SCNu32, &a_height);
		sscanf(argv[3], "%" SCNu8, &a_countries);
		sscanf(argv[4], "%lf", &a_watertoland);
		sscanf(argv[5], "%" SCNu32, &a_population);
		sscanf(argv[6], "%lf", &a_travelchance);
		sscanf(argv[7], "%lf", &a_contagiousness);
		sscanf(argv[8], "%lf", &a_deadliness);
		sscanf(argv[9], "%lf", &a_curability);

		debug_print("%" PRIu32 " %" PRIu32 " %" PRIu8 " %lf %" PRIu32 " %lf %lf %lf %lf", a_width, a_height, a_countries, a_watertoland, a_population, a_travelchance, a_contagiousness, a_deadliness, a_curability);
		return 0;
	}
}

void print_ppm(AREA *myworld) {
	FILE *fp;
	float att;

	/* buffer[0] = red, buffer[1] = green, buffer[2] = blue */
	unsigned char buffer[3 * a_width * a_height];
	memset(&buffer[0], 0, sizeof(buffer));

	char filename[255];
	AREA area;

	sprintf(filename, "world%05d.ppm", frame);
	/* ppm header */
	fp = fopen(filename, "w");
	fprintf(fp, "P6\n%" PRIu32 " %" PRIu32 "\n255\n", a_width, a_height);
	for (int i = 0; i < a_width * a_height; i++) {
		area = myworld[i];
		if (!area.type) {
			buffer[3*i] = 255;
			buffer[3*i + 2] = 255;
			continue;
		} else if (area.type == WATER) {
			/* water */
			buffer[3*i + 2] = 128;
			continue;
		} else if (area.infected) {
			/* some infected */
			buffer[3*i + 0] = 255;
			buffer[3*i + 1] = 255 * ((float) (area.alive - area.infected)) / area.alive;
		} else if (area.alive) {
			/* no infected */
			buffer[3*i + 1] = 255;
		} else {
			buffer[3*i] = 0;
			buffer[3*i + 1] = 0;
			buffer[3*i + 2] = 0;
			continue;
		}
		/* deadites */
		if (area.dead && area.alive) {
			if (area.dead > area.alive) {
				att = (float) area.alive / area.dead;
				//~ debug_print("lo: %f", att);
			} else {
				att = (float) area.alive / (area.dead + area.alive);
				//~ debug_print("hi: %f", att);
			}
			buffer[3*i] = buffer[3*i] * att;
			buffer[3*i + 1] = buffer[3*i + 1] * att;
			buffer[3*i + 2] = 0;
		}
	}
	fwrite(&buffer[0], 1, sizeof(buffer), fp);
	fclose(fp);
	frame++;
}

void print_text(AREA *myworld) {
	for (int i = 0; i < a_width * a_height; i++) {
		putchar((int) '0' + myworld[i].type);
		if (!((i+1) % a_width)) {
			putchar(10);
		}
	}
}

int make_world(void) {
	uint64_t x;
	uint64_t pos;
	bool cells = true;
	uint64_t size = a_width * a_height;
	uint8_t mytype;

	/* land */
	for (int i = 2; i < 2+a_countries; i++) {
		do {
			x = rand() % size;
		} while (world[w1][x].type);

		// TODO: create areas with random population
		world[w1][x].type = i;
		world[w1][x].alive = a_population;

		// TODO: actually create single persons
		//~ world[w1][x].people = (PERSON *) malloc(a_population * sizeof(PERSON));
	}

	/* water */
	for (int i = 0; i < (a_watertoland * a_countries); i++) {
		do {
			x = rand() % size;
		} while (world[w1][x].type);
		world[w1][x].type = WATER;
	}

	/* grow the areas */
	while (cells) {
		cells = false;
		for (x = 0; x < size; x++) {
			mytype = world[w1][x].type;

			if(mytype) {

				// copy current cell to w2
				world[w2][x].type = mytype;
				world[w2][x].alive = world[w1][x].alive;

				// TODO: better coastlines
				if (!(rand() % 3)) {
					continue;
				}

				// above
				pos = (size + x - a_width) % size;
				if (!world[w1][pos].type) {
					cells = true;
					world[w2][pos].type = mytype;
					world[w2][pos].alive = a_population;
				}

				// below
				pos = (x + a_width) % size;
				if (!world[w1][pos].type) {
					cells = true;
					world[w2][pos].type = mytype;
					world[w2][pos].alive = a_population;
				}

				// left
				pos = (size + x - 1) % size;
				if (!world[w1][pos].type) {
					cells = true;
					world[w2][pos].type = mytype;
					world[w2][pos].alive = a_population;
				}

				// right
				pos = (x + 1) % size;
				if (!world[w1][pos].type) {
					cells = true;
					world[w2][pos].type = mytype;
					world[w2][pos].alive = a_population;
				}
			}
		}

		//~ print_ppm(world[w1]);
		/* swap worlds */
		uint8_t i = w1; w1 = w2; w2 = i;
	}
	//~ free(world[w2]);
	return w1;
}

void init_world(void) {
	/* TODO: check for success */
	world = (AREA **) calloc(2, sizeof(AREA *));
	world[0] = (AREA *) calloc(a_width * a_height, sizeof(AREA));
	world[1] = (AREA *) calloc(a_width * a_height, sizeof(AREA));
}

void free_worlds(void) {
	// FIXME: doesn't actually free everything
	//~ for (int i = 0; i < a_width * a_height; i++) {
		//~ if (world[0][i].people) {
			//~ free(world[0][i].people);
		//~ }
		//~ if (world[1][i].people) {
			//~ free(world[1][i].people);
		//~ }
	//~ }
	free(world[0]);
	free(world[1]);
	free(world);
}

int run_simulation(AREA *myworld) {
	uint64_t size = a_width * a_height;
	uint64_t src;
	uint64_t dst;
	uint64_t chance;
	uint8_t mytype;
	uint64_t travellers = 0;
	uint64_t sick;
	uint64_t mysick;
	uint64_t myalive = 0;
	uint64_t allalive = 0;
	uint64_t allsick = 0;

	for (src = 0; src < size; src++) {
		travellers = 0;
		mytype = myworld[src].type;
		if (mytype == WATER) {
			continue;
		}

		myalive = myworld[src].alive;

		if (myalive > 0) {
			if (myworld[src].infected) {
				// heal me, maybe?
				mysick = myworld[src].infected;
				for (int i = 0; i < mysick; i++) {
					if (uniform_deviate(rand()) < a_curability) {
						myworld[src].infected--;
					}
				}

				myworld[src].infected *= a_contagiousness;
				// spread to one other guy if there's just one who's sick
				if (myworld[src].infected == 1 && a_contagiousness < 2.0f) {
					myworld[src].infected = 2;
				}

				// no sick zombies
				if (myworld[src].infected > myalive) {
					myworld[src].infected = myalive;
				}

				// kill some
				mysick = myworld[src].infected;
				for (int i = 0; i < mysick; i++) {
					if (uniform_deviate(rand()) < a_deadliness) {
						myworld[src].infected--;
						myworld[src].alive--;
						myworld[src].dead++;
					}
				}
			}

			allsick += myworld[src].infected;
			allalive += myworld[src].alive;

			chance = uniform_deviate(rand()) * 12;
			//~ debug_print("chance: %" PRIu64, chance);
			if (chance < 2) {
				// above
				dst = (size + src - a_width) % size;
			} else if (chance < 4) {
				// above
				dst = (src + a_width) % size;
			} else if (chance < 6) {
				// left
				dst = (size + src - 1) % size;
			} else if (chance < 8) {
				// right
				dst = (src + 1) % size;
			} else if (chance < 10) {
				do {
					dst = uniform_deviate(rand()) * (size);
				} while (myworld[dst].type != mytype);
			} else {
				do {
					dst = uniform_deviate(rand()) * (size);
				} while (myworld[dst].type == WATER);
			}

			if (myworld[dst].type == WATER) {
				continue;
			}

			for (uint64_t i = 0; i < myworld[src].alive; i++) {
				if (uniform_deviate(rand()) < a_travelchance) {
					travellers++;
					//~ debug_print("HELLO, THIS IS %" PRIu64" at PERSON %" PRIu64, src, i);
				}
			}

			if (!travellers) {
				continue;
			}
			//~ debug_print("travellers: %" PRIu64, travellers);
			sick = myworld[src].infected * (((float) travellers) / myalive);
			myworld[src].infected -= sick;
			myworld[src].alive -= travellers;
			myworld[dst].infected += sick;
			myworld[dst].alive += travellers;
		}
	}
	debug_print("still %" PRIu64 " sick people, %" PRIu64 " alive", allsick, allalive);
	return allsick;
}

void infect_person(AREA *myworld) {
	uint64_t x;
	do {
		//~ x = rand() % (a_width * a_height);
		x = uniform_deviate(rand()) * (a_width * a_height);
	} while (myworld[x].type == WATER);
	myworld[x].infected = INDEXCASES;
	debug_print("initial: %" PRIu64, x);
}

int main(int argc, char **argv) {
	uint64_t proc = 1;
	srand(time_seed());

		// parse arguments
	if (parseargs(argc, argv)) {
		fprintf(stderr, "could not parse args\n");
		return 1;
	}

	init_world();
	w = world[make_world()];
	// infect somebody
	for (int i=0; i < 1; i++) {
		infect_person(w);
	}
	while(proc) {
		proc = run_simulation(w);
		print_ppm(w);
	}
	//~ stuff
	free_worlds();
	return 0;
}
